<?php

namespace EventPortal\Domain\ActionList;

use EventPortal\Domain\ActionList\Models\ActionList as ActionListModel;
use EventPortal\Domain\ActionList\Models\Task as TaskModel;
use EventPortal\Domain\Event\Models\Event;
use EventPortal\Domain\User\Models\User;
use Illuminate\Support\Str;

class Task
{
    public ?string $name = null;

    public string $label;

    public bool $required = false;

    public bool $automatic = false;

    public ?string $route = null;

    public ?array $routeParams = null;

    public ?string $href = null;

    public ?string $actionLabel = null;

    public ?Event $event = null;

    public function __construct(?string $name = null)
    {
        $this->name = $name;
    }

    public static function make(?string $name = null): self
    {
        return new static($name);
    }

    public function label(string $label): self
    {
        $this->label = $label;

        if ($this->name === null) {
            $this->name = Str::slug($label, '_');
        }

        return $this;
    }

    public function required(): self
    {
        $this->required = true;

        return $this;
    }

    public function automatic(): self
    {
        $this->automatic = true;

        return $this;
    }

    public function route(string $route, ?array $routeParams = [], ?string $label = null): self
    {
        $this->route = $route;
        $this->routeParams = $routeParams;
        $this->actionLabel = $label;

        return $this;
    }

    public function href(string $href, ?string $label = null): self
    {
        $this->href = $href;
        $this->actionLabel = $label;

        return $this;
    }

    public function actionLabel(string $label): self
    {
        $this->actionLabel = $label;

        return $this;
    }

    public function event(Event $event): self
    {
        $this->event = $event;

        return $this;
    }

    public function complete(User $user): void
    {
        $actionList = ActionListModel::where('event_id', $this->event->id)
            ->where('type', $this->event->typeValue)
            ->where('status', $this->event->status)
            ->first();

        if ($actionList === null) {
            return;
        }

        $task = $actionList->tasks()->where('name', $this->name)->first();

        if ($task === null) {
            return;
        }

        $task->setAsComplete($user);
    }

    public function getAction(): ?array
    {
        if ($this->route !== null) {
            return [
                'type' => 'route',
                'route' => $this->route,
                'routeParams' => $this->routeParams,
                'label' => $this->actionLabel,
            ];
        }

        if ($this->href !== null) {
            return [
                'type' => 'href',
                'route' => $this->href,
                'label' => $this->actionLabel,
            ];
        }

        return null;
    }

    public function get(): array
    {
        if ($this->event === null) {
            return [];
        }

        $actionList = ActionListModel::where('event_id', $this->event->id)
            ->where('type', $this->event->typeValue)
            ->where('status', $this->event->status)
            ->first();

        if ($actionList === null) {
            return [];
        }

        return $actionList
            ->tasks
            ->map(function (TaskModel $task) {
                return [
                    'id' => $task->id,
                    'label' => $task->label,
                    'action' => $task->action,
                    'completed' => (bool) $task->completed_at,
                    'automatic' => (bool) $task->automatic,
                ];
            })
            ->toArray();
    }
}
