<?php

namespace EventPortal\Tests\Unit\Form;

use Closure;
use EventPortal\Domain\Component\Button;
use EventPortal\Domain\Component\Input\Text;
use EventPortal\Domain\Component\Section;
use EventPortal\Domain\Delegate\Models\Delegate;
use EventPortal\Domain\Event\Form\Enums\QuestionType;
use EventPortal\Domain\Event\Form\Models\Form as FormModel;
use EventPortal\Domain\Event\Form\Models\FormStep;
use EventPortal\Domain\Event\Form\Models\FormStepSection;
use EventPortal\Domain\Event\Form\Models\Question;
use EventPortal\Domain\Event\Form\Models\QuestionOption;
use EventPortal\Domain\Event\Form\Models\QuestionValidation;
use EventPortal\Domain\Event\Models\Event;
use EventPortal\Domain\Form\Facades\Form;
use EventPortal\Domain\Handlers\UpdateModel;
use EventPortal\Tests\Support\TestModel;
use EventPortal\Tests\Support\TestModelUpdateHandler;
use EventPortal\Tests\TestCase;
use Illuminate\Foundation\Testing\RefreshDatabase;
use Illuminate\Support\Arr;
use Illuminate\Validation\ValidationException;
use Spatie\Snapshots\MatchesSnapshots;

class FormTest extends TestCase
{
    use RefreshDatabase, MatchesSnapshots;

    /** @test **/
    public function makes_instance()
    {
        $form = Form::make('test.create');

        $this->assertInstanceOf(\EventPortal\Domain\Form\Form::class, $form);
    }

    /** @test **/
    public function can_make_from_database(): void
    {
        $formModel = factory(FormModel::class)->create();
        factory(FormStep::class)->create(['form_id' => $formModel->id]);
        $form = Form::makeFromDatabase($formModel);

        $this->assertInstanceOf(\EventPortal\Domain\Form\Form::class, $form);
    }

    /** @test **/
    public function makes_mutiple_instances()
    {
        $formA = Form::make('test.create');
        $formB = Form::make('test.edit');

        $this->assertNotSame($formA, $formB);
        $this->assertEquals('test.create', $formA->name);
        $this->assertEquals('test.edit', $formB->name);
    }

    /** @test **/
    public function can_add_array_of_fields()
    {
        $form = Form::make('test.create')
            ->addFields(function () {
                return [
                    Section::make('Key Details')->fields([
                        Text::make('Name'),
                    ]),
                ];
            });

        $this->assertIsArray($form->blocks);
        $this->assertCount(1, $form->blocks);
    }

    /** @test **/
    public function can_call_add_fields_multiple_times()
    {
        $form = Form::make('test.create')
            ->addFields(function () {
                return [
                    Section::make('Key Details')->fields([
                        Text::make('Name'),
                    ]),
                ];
            })
            ->addFields(function () {
                return [
                    Text::make('Email'),
                ];
            });

        $this->assertIsArray($form->blocks);
        $this->assertCount(2, $form->blocks);
    }

    /** @test **/
    public function field_without_section_creates_invisible_section()
    {
        $form = Form::make('test.create')
            ->model(new Event)
            ->addFields(function () {
                return [
                    Text::make('Email'),
                ];
            })
            ->render();

        $this->assertInstanceOf(Section::class, head($form['content'][0]->children));
    }

    /** @test **/
    public function can_add_fields_with_a_callable()
    {
        $form = Form::make('test.create')->addFields(function () {
            return [
                Text::make('name'),
            ];
        });

        $this->assertTrue(head($form->blocks) instanceof Closure);
    }

    /** @test **/
    public function can_add_action()
    {
        $form = Form::make('test.create')
            ->addAction(function () {
                return Button::make('Test Button');
            })
            ->render();

        $this->assertTrue(is_array($form['actions']));
        $this->assertInstanceOf(Button::class, $form['actions'][0]);
    }

    /** @test **/
    public function can_add_primary_action_with_a_callable()
    {
        $form = Form::make('test.create')
            ->addPrimaryAction(function () {
                return true;
            });

        $this->assertTrue($form->primaryAction instanceof Closure);
    }

    /** @test */
    public function can_resolve_primary_action()
    {
        $form = Form::make('test.edit')
            ->model(new Event)
            ->addPrimaryAction(function () {
                return 'Cheese';
            })
            ->render();

        $this->assertEquals('Cheese', $form['onEnter']);
    }

    /** @test */
    public function can_resolve_with_context_primary_action()
    {
        $event = factory(Event::class)->create();
        $form = Form::make('test.edit')
            ->model($event)
            ->addPrimaryAction(function (Event $event) {
                return $event->name;
            })
            ->render();

        $this->assertEquals($event->name, $form['onEnter']);
    }

    /** @test **/
    public function reset_after_handle()
    {
        $event = factory(Event::class)->create(['name' => 'Old Name']);

        Form::make('test.edit')->addHandler(UpdateModel::class);

        Form::make('test.edit')
            ->model($event)
            ->handle([
                'name' => 'New Name',
            ]);

        $newFormInstance = Form::make('test.edit');

        $this->assertEquals([], $newFormInstance->blocks);
        $this->assertEquals([], $newFormInstance->actions);
        $this->assertNull($newFormInstance->primaryAction);

        $this->assertDatabaseHas('events', ['name' => 'New Name']);
    }

    /** @test */
    public function can_render_form_from_database(): void
    {
        $formModel = factory(FormModel::class)->create();
        $step = factory(FormStep::class)->state('snapshot')->create(['form_id' => $formModel->id]);
        $section = factory(FormStepSection::class)->state('snapshot')->create(['form_step_id' => $step->id]);
        $question = factory(Question::class)->states(['snapshot', 'gdpr'])
            ->create(['type' => QuestionType::MultipleChoiceAnswers(), 'form_step_section_id' => $section->id]);
        factory(QuestionOption::class)->state('snapshot')->create(['question_id' => $question->id, 'value' => true]);
        factory(QuestionOption::class)->state('snapshot')->create(['question_id' => $question->id, 'value' => true]);
        factory(QuestionValidation::class)->state('required')->create(['question_id' => $question->id]);

        $form = Form::makeFromDatabase($formModel)->model(new Delegate);

        $this->assertMatchesJsonSnapshot(collect($form->render())->toJson());
    }

    /** @test */
    public function can_render_form_from_database_in_the_correct_order(): void
    {
        $formModel = factory(FormModel::class)->create();
        $steps = $this->createReverseOrdered(FormStep::class, 'form_id', [$formModel]);
        $sections = $this->createReverseOrdered(FormStepSection::class, 'form_step_id', $steps);
        $questions = $this->createReverseOrdered(
            Question::class,
            'form_step_section_id',
            $sections,
            ['type' => QuestionType::MultipleChoiceAnswers()]
        );
        $this->createReverseOrdered(QuestionOption::class, 'question_id', $questions, ['value' => true]);

        $form = Form::makeFromDatabase($formModel)->model(new Delegate);
        $this->assertMatchesJsonSnapshot(collect($form->render())->toJson());
    }

    /** @test **/
    public function handlers_are_invoked()
    {
        $model = factory(TestModel::class)->create(['name' => 'Jeff']);

        Form::make('test.create')
            ->addHandler(TestModelUpdateHandler::class)
            ->model($model)
            ->handle([
                'name' => 'Lee',
            ]);

        $this->assertEquals('Lee', TestModel::first()->name);
    }

    /** @test **/
    public function handlers_are_not_newed_twice()
    {
        $model = factory(TestModel::class)->create(['name' => 'Jeff']);

        Form::make('test.create')
            ->addHandler(TestModelUpdateHandler::class)
            ->model($model)
            ->handle([
                'name' => 'Lee',
            ]);

        Form::make('test.create')
            ->model($model)
            ->handle([
                'name' => 'Mike',
            ]);

        $this->assertEquals('Mike', TestModel::first()->name);
    }

    /** @test */
    public function throws_toast_when_form_validation_is_fails(): void
    {
        try {
            Form::make('test.create')
                ->addFields(function () {
                    return [
                        Text::make('name')->rules(['required']),
                    ];
                })
                ->validate(['name' => '']);
        } catch (ValidationException $e) {
            $this->assertTrue(session()->has('toast'));
            $this->assertEquals(session('toast'), [
                'type' => 'error',
                'message' => 'There has been a problem saving the form. Please review and correct any errors.',
                'timeout' => 0,
            ]);
        }
    }

    private function createReverseOrdered(string $type, string $parentId, array $children, array $overrides = []): array
    {
        $result = [];
        $name = $type === Question::class ? 'question' : 'name';
        foreach ($children as $child) {
            array_push(
                $result,
                factory($type)->state('snapshot')
                    ->create(array_merge([
                        $name => $this->getClassName($type) . 2,
                        $parentId => $child->id,
                        'order' => 2,
                    ], $overrides))
            );
            array_push(
                $result,
                factory($type)->state('snapshot')
                    ->create(array_merge([
                        $name => $this->getClassName($type) . 1,
                        $parentId => $child->id,
                        'order' => 1,
                    ], $overrides))
            );
        }

        return $result;
    }

    private function getClassName(string $class): string
    {
        return Arr::last(explode('\\', $class));
    }
}
